@@ -440,11 +440,9 @@ class FTPFS(FS):
     def ftp_url(self):
         # type: () -> Text
         """Get the FTP url this filesystem will open."""
-        url = (
-            "ftp://{}".format(self.host)
-            if self.port == 21
-            else "ftp://{}:{}".format(self.host, self.port)
-        )
+        _host_part = self.host if self.port == 21 else "{}:{}".format(self.host, self.port)
+        _user_part = "" if self.user == "anonymous" or self.user is None else "{}:{}@".format(self.user, self.passwd)
+        url = "ftp://{}{}".format(_user_part, _host_part)
         return url
 
     @property
@@ -454,6 +452,14 @@ class FTPFS(FS):
         """
         return self._get_ftp()
 
+    def geturl(self, path, purpose="download"):
+        # type: (str, str) -> Text
+        """Get FTP url for resource."""
+        _path = self.validatepath(path)
+        if purpose != "download":
+            raise errors.NoURL(_path, purpose)
+        return "{}{}".format(self.ftp_url, _path)
+
     def _get_ftp(self):
         # type: () -> FTP
         if self._ftp is None:
