@@ -19,15 +19,18 @@ the expense of potentially copying extra files.
 from __future__ import print_function
 from __future__ import unicode_literals
 
+from contextlib import contextmanager
 import typing
 
+from ._bulk import Copier
 from .copy import copy_file_internal
 from .errors import ResourceNotFound
-from .walk import Walker
 from .opener import manage_fs
+from .tools import is_thread_safe
+from .walk import Walker
 
 if False:  # typing.TYPE_CHECKING
-    from typing import Optional, Text, Union
+    from typing import Callable, Optional, Text, Union
     from .base import FS
     from .info import Info
 
@@ -49,11 +52,13 @@ def _compare(info1, info2):
     return date1 is None or date2 is None or date1 > date2
 
 
-def mirror(src_fs,              # type: Union[FS, Text]
-           dst_fs,              # type: Union[FS, Text]
-           walker=None,         # type: Optional[Walker]
-           copy_if_newer=True   # type: bool
-           ):
+def mirror(
+    src_fs,  # type: Union[FS, Text]
+    dst_fs,  # type: Union[FS, Text]
+    walker=None,  # type: Optional[Walker]
+    copy_if_newer=True,  # type: bool
+    workers=0,  # type: int
+):
     # type: (...) -> None
     """Mirror files / directories from one filesystem to another.
 
@@ -66,28 +71,40 @@ def mirror(src_fs,              # type: Union[FS, Text]
         dst_fs (FS or str): Destination filesystem (URL or instance).
         walker (~fs.walk.Walker, optional): An optional walker instance.
         copy_if_newer (bool): Only copy newer files (the default).
-
+        workers (int): Number of worker threads used
+            (0 for single threaded). Set to a relatively low number
+            for network filesystems, 4 would be a good start.
     """
-    with manage_fs(src_fs, writeable=False) as _src_fs:
-        with manage_fs(dst_fs, create=True) as _dst_fs:
-            with _src_fs.lock(), _dst_fs.lock():
-                return _mirror(
+
+    def src():
+        return manage_fs(src_fs, writeable=False)
+
+    def dst():
+        return manage_fs(dst_fs, create=True)
+
+    with src() as _src_fs, dst() as _dst_fs:
+        with _src_fs.lock(), _dst_fs.lock():
+            _thread_safe = is_thread_safe(_src_fs, _dst_fs)
+            with Copier(num_workers=workers if _thread_safe else 0) as copier:
+                _mirror(
                     _src_fs,
                     _dst_fs,
                     walker=walker,
-                    copy_if_newer=copy_if_newer
+                    copy_if_newer=copy_if_newer,
+                    copy_file=copier.copy,
                 )
 
 
-def _mirror(src_fs, dst_fs, walker=None, copy_if_newer=True):
-    # type: (FS, FS, Optional[Walker], bool) -> None
+def _mirror(
+    src_fs, dst_fs, walker=None, copy_if_newer=True, copy_file=copy_file_internal
+):
+    # type: (FS, FS, Optional[Walker], bool, Callable[[FS, str, FS, str], None]) -> None
     walker = walker or Walker()
-    walk = walker.walk(src_fs, namespaces=['details'])
+    walk = walker.walk(src_fs, namespaces=["details"])
     for path, dirs, files in walk:
         try:
             dst = {
-                info.name: info
-                for info in dst_fs.scandir(path, namespaces=['details'])
+                info.name: info for info in dst_fs.scandir(path, namespaces=["details"])
             }
         except ResourceNotFound:
             dst_fs.makedir(path)
@@ -105,7 +122,7 @@ def _mirror(src_fs, dst_fs, walker=None, copy_if_newer=True):
                     # Compare file info
                     if copy_if_newer and not _compare(_file, dst_file):
                         continue
-            copy_file_internal(src_fs, _path, dst_fs, _path)
+            copy_file(src_fs, _path, dst_fs, _path)
 
         # Make directories
         for _dir in dirs:
